# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for Node."""
from functools import wraps
from typing import (
    Any,
    Callable,
    Tuple,
)

import forge
import pythonflow as pf

from qctrlcommons.exceptions import QctrlException


class _NumpyCompatibleFuncOp(pf.func_op):
    """
    Version of `pf.func_op` that can be used in binary operations with NumPy arrays.

    By default ``arr + pf.func_op(*)`` throws an error, since NumPy doesn't know how to add
    `pf.func_op` objects to arrays. Even the fact that the func ops override `__radd__` doesn't
    help, since the NumPy addition takes precedence. We can instead tell NumPy to delegate all
    binary operations to the other operand, by explicitly clearing the `__array_ufunc__` attribute.
    """

    __array_ufunc__ = None


def _validate_kwargs(func: Callable, func_kwargs: list):
    """
    Decorator for validating the function's keyword arguments.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        for kwarg in kwargs:
            if kwarg not in func_kwargs:
                raise QctrlException(
                    f"{func.__name__}() got an unexpected keyword argument '{kwarg}'"
                )
        return func(*args, **kwargs)

    return wrapper


def _numpy_compatible_opmethod(target: Callable):
    """
    Version of `pf.opmethod` that returns instances of `_numpy_compatible_func_op`.
    """

    @wraps(target)
    def _func(*args, **kwargs):
        return _NumpyCompatibleFuncOp(target, *args, **kwargs)

    return _func


class Node:
    """Custom structure."""

    name = None
    args = []
    rtype = Any

    def __init__(self, node_id, args, kwargs):
        self.node_id = node_id
        self._args = args
        self._kwargs = kwargs

    @classmethod
    def valid_kwargs(cls) -> list:
        """
        Returns the list of valid kwargs.

        Returns
        -------
        List[str]
            The list of valid kwargs for this function.
        """
        pythonflow_func_kwargs = ["name"]
        custom_func_kwargs = [str(arg.name) for arg in cls.args]
        return custom_func_kwargs + pythonflow_func_kwargs

    @classmethod
    def create_pf(cls):
        """Creates the corresponding pythonflow function."""

        def func(*args, **kwargs):
            raise RuntimeError("cannot execute locally")

        func.__doc__ = cls.__doc__
        func.__name__ = cls.name  # pylint:disable=non-str-assignment-to-dunder-name
        sig = forge.sign(*cls.args)
        func = sig(func)
        func = forge.returns(cls.rtype)(func)
        func = _numpy_compatible_opmethod(func)
        func = _validate_kwargs(func=func, func_kwargs=cls.valid_kwargs())

        return func

    def _evaluate_inputs(
        self, execution_context: "ExecutionContext"
    ) -> Tuple[Any, Any]:
        """Evaluates the args and kwargs. Called before evaluating the node
        itself.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the node value.

        Returns
        -------
        Tuple[Any, Any]
            the value of args, kwargs
        """
        args = []

        for arg in self._args:
            value = arg
            if isinstance(arg, Node):
                value = arg.evaluate(execution_context)
            elif isinstance(arg, list):
                value = [node.evaluate(execution_context) for node in arg]
            elif isinstance(arg, dict):
                value = {key: arg[key].evaluate(execution_context) for key in arg}

            args.append(value)

        kwargs = {}

        for key, value in self._kwargs.items():

            evaluated_value = value
            if isinstance(value, Node):
                evaluated_value = value.evaluate(execution_context)
            elif isinstance(value, list):
                evaluated_value = [node.evaluate(execution_context) for node in value]
            elif isinstance(value, dict):
                evaluated_value = {
                    key: value[key].evaluate(execution_context) for key in value
                }

            kwargs[key] = evaluated_value

        return args, kwargs

    def _evaluate_node(self, execution_context, args, kwargs):
        """Given evaluated args and kwargs, return the evaluated node value.

        Parameters
        ----------
        execution_context :
            helper class for evaluating the node value.
        args
            argument list for the node.
        kwargs
            keyword arguments for the node.

        """
        raise NotImplementedError

    def evaluate(self, execution_context: "ExecutionContext") -> Any:
        """Get the value from the Node itself.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the value of node.

        Returns
        -------
        Any
            the value of the node.
        """
        # check if it is evaluated
        if execution_context.is_recorded(self):
            return execution_context.get_recorded_value(self)

        args, kwargs = self._evaluate_inputs(execution_context)
        value = self._evaluate_node(execution_context, args, kwargs)
        execution_context.record(self, value)
        return value
