# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for CoreNode."""
from numbers import Number
from typing import Union

import forge

from qctrlcommons.node.module import ModuleNode

from . import types


class TensorFlowNode(ModuleNode):
    """Represents CoreNode class."""

    _module_name = "tensorflow"

    def _evaluate_node(self, execution_context, args, kwargs):
        """Handle different dtype of value before evaluate the node.

        Parameters
        ----------
        execution_context : ExecutionContext
            helper class for evaluating the value of the node.
        args
            argument list for the node.
        kwargs
            keyword arguments for the node.

        Returns
        -------
        Any
            the value of the node.

        """
        tf = execution_context.tensorflow  # pylint:disable=invalid-name

        # if one of args is tensor,
        # we need to get the dtype value,
        cast_type = None
        for arg in args:
            if tf.is_tensor(arg):
                cast_type = arg.dtype

        for index, arg in enumerate(args):
            # create the tensor with tf.constant and
            # cast the tensor to the same dtype
            if isinstance(arg, Number) and cast_type:
                args[index] = tf.cast(tf.constant(arg), dtype=cast_type)

        return super()._evaluate_node(execution_context, args, kwargs)


class Truediv(TensorFlowNode):
    """
    Divides two values element-wise.

    Parameters
    ----------
    x : number or Tensor
        The numerator, :math:`x`.
    y : number or Tensor
        The denominator, :math:`y`.

    Returns
    -------
    Tensor
        The element-wise value of :math:`x/y`.
    """

    name = "truediv"
    _module_attr = "truediv"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
        forge.arg("y", type=Union[float, complex, types.Tensor]),
    ]


class Multiply(TensorFlowNode):
    r"""
    Multiplies two values element-wise.

    Parameters
    ----------
    x : number or Tensor
        The left multiplicand, :math:`x`.
    y : number or Tensor
        The right multiplicand, :math:`y`.

    Returns
    -------
    Tensor
        The element-wise value of :math:`x\times y`.
    """

    name = "mul"
    _module_attr = "multiply"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
        forge.arg("y", type=Union[float, complex, types.Tensor]),
    ]


class Addition(TensorFlowNode):
    """
    Adds two values element-wise.

    Parameters
    ----------
    x : number or Tensor
        The left summand, :math:`x`.
    y : number or Tensor
        The right summand, :math:`y`.

    Returns
    -------
    Tensor
        The element-wise value of :math:`x+y`.
    """

    name = "add"
    _module_attr = "add"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
        forge.arg("y", type=Union[float, complex, types.Tensor]),
    ]


class Subtract(TensorFlowNode):
    """
    Subtracts one value from another element-wise.

    Parameters
    ----------
    x : number or Tensor
        The value from which to subtract, :math:`x`.
    y : number or Tensor
        The value to subtract, :math:`y`.

    Returns
    -------
    Tensor
        The element-wise value of :math:`x-y`.
    """

    name = "sub"
    _module_attr = "subtract"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
        forge.arg("y", type=Union[float, complex, types.Tensor]),
    ]


class Absolute(TensorFlowNode):
    r"""
    Returns the element-wise absolute value.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise absolute value :math:`\left|x\right|`.
    """

    name = "abs"
    _module_attr = "abs"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Exponent(TensorFlowNode):
    """
    Raises one value to the power of another element-wise.

    Parameters
    ----------
    x : number or Tensor
        The base, :math:`x`.
    y : number or Tensor
        The exponent, :math:`y`.

    Returns
    -------
    Tensor
        The element-wise value of :math:`x^y`.
    """

    name = "pow"
    _module_attr = "pow"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
        forge.arg("y", type=Union[float, complex, types.Tensor]),
    ]


class FloorDivision(TensorFlowNode):
    r"""
    Divides two values and then takes the floor of the result, element-wise.

    Parameters
    ----------
    x : number or Tensor
        The numerator, :math:`x`.
    y : number or Tensor
        The denominator, :math:`y`.

    Returns
    -------
    Tensor
        The element-wise value of :math:`\lfloor x/y \rfloor`.
    """

    name = "floordiv"
    _module_attr = "floordiv"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
        forge.arg("y", type=Union[float, complex, types.Tensor]),
    ]


class ComplexValue(TensorFlowNode):
    """
    Creates a complex number from real and imaginary parts.

    Parameters
    ----------
    real : float or Tensor
        The real part, :math:`a`.
    imag : float or Tensor
        The imaginary part, :math:`b`.

    Returns
    -------
    Tensor
        The complex number :math:`a+ib` element-wise.
    """

    name = "complex_value"
    _module_attr = "complex"
    args = [
        forge.arg("real", type=Union[float, types.Tensor]),
        forge.arg("imag", type=Union[float, types.Tensor]),
    ]


class Sqrt(TensorFlowNode):
    r"""
    Returns the element-wise square root.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise square root :math:`\sqrt{x}`.
    """

    name = "sqrt"
    _module_attr = "sqrt"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Exp(TensorFlowNode):
    r"""
    Returns the element-wise exponential.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise exponential :math:`\exp{x}`.
    """

    name = "exp"
    _module_attr = "exp"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Sin(TensorFlowNode):
    r"""
    Returns the element-wise sine.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise sine :math:`\sin{x}`.
    """

    name = "sin"
    _module_attr = "sin"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Cos(TensorFlowNode):
    r"""
    Returns the element-wise cosine.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise cosine :math:`\cos{x}`.
    """

    name = "cos"
    _module_attr = "cos"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Tan(TensorFlowNode):
    r"""
    Returns the element-wise tangent.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise tangent :math:`\tan{x}`.
    """

    name = "tan"
    _module_attr = "tan"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Arcsin(TensorFlowNode):
    r"""
    Returns the element-wise arcsine.

    Parameters
    ----------
    x : float or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise arcsine :math:`\arcsin{x}`.
        Outputs will be in the range of :math:`[-\pi/2, \pi/2]`.
    """

    name = "arcsin"
    _module_attr = "asin"
    args = [
        forge.arg("x", type=Union[float, types.Tensor]),
    ]


class Arccos(TensorFlowNode):
    r"""
    Returns the element-wise arccosine.

    Parameters
    ----------
    x : float or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise arccosine :math:`\arccos{x}`.
        Outputs will be in the range of :math:`[0, \pi]`.
    """

    name = "arccos"
    _module_attr = "acos"
    args = [
        forge.arg("x", type=Union[float, types.Tensor]),
    ]


class Arctan(TensorFlowNode):
    r"""
    Returns the element-wise arctangent.

    Parameters
    ----------
    x : float or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise arctangent :math:`\arctan{x}`.
        Outputs will be in the range of :math:`(-\pi/2, \pi/2)`.
    """

    name = "arctan"
    _module_attr = "atan"
    args = [
        forge.arg("x", type=Union[float, types.Tensor]),
    ]


class Sum(TensorFlowNode):
    r"""
    Returns the sum of all the elements in a tensor.

    Parameters
    ----------
    input_tensor : Tensor
        The tensor whose elements you want to sum.

    Returns
    -------
    Tensor
        The single-element tensor containing the input tensor's sum.
    """

    name = "sum"
    _module_attr = "reduce_sum"
    args = [
        forge.arg("input_tensor", type=types.Tensor),
    ]


class Log(TensorFlowNode):
    r"""
    Returns the element-wise natural logarithm.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise natural logarithm :math:`\ln{x}`.
    """

    name = "log"
    _module_attr = "math.log"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Cosh(TensorFlowNode):
    r"""
    Returns the element-wise hyperbolic cosine.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise hyperbolic cosine :math:`\cosh{x}`.
    """

    name = "cosh"
    _module_attr = "cosh"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Sinh(TensorFlowNode):
    r"""
    Returns the element-wise hyperbolic sine.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise hyperbolic sine :math:`\sinh{x}`.
    """

    name = "sinh"
    _module_attr = "sinh"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Tanh(TensorFlowNode):
    r"""
    Returns the element-wise hyperbolic tangent.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise hyperbolic tangent :math:`\tanh{x}`.
    """

    name = "tanh"
    _module_attr = "tanh"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]


class Negative(TensorFlowNode):
    r"""
    Returns element-wise negative value.

    Parameters
    ----------
    x : number or Tensor
        The value, :math:`x`.

    Returns
    -------
    Tensor
        The element-wise negative value :math:`-x`.
    """

    name = "neg"
    _module_attr = "negative"
    args = [
        forge.arg("x", type=Union[float, complex, types.Tensor]),
    ]
